/**
 * 
 */
package org.apache.activemq.test;

import java.util.Enumeration;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;

import javax.jms.Connection;
import javax.jms.Destination;
import javax.jms.JMSException;
import javax.jms.Queue;
import javax.jms.QueueBrowser;
import javax.jms.Session;
import javax.management.MBeanServerConnection;
import javax.management.ObjectName;
import javax.management.remote.JMXConnector;
import javax.management.remote.JMXConnectorFactory;
import javax.management.remote.JMXServiceURL;

import org.apache.activemq.ActiveMQConnectionFactory;
import org.apache.activemq.broker.BrokerService;

import org.slf4j.LoggerFactory;
import org.slf4j.Logger;
import org.apache.xbean.spring.context.ClassPathXmlApplicationContext;

import org.junit.After;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;

/**
 * Unit test for bug
 */
public class AMQTemplateTest {

    //constants
    protected final static String CONFIG_FILE="AMQTemplateTest.xml";
    protected final static String BROKERURL = "tcp://localhost:61616";
    protected final static String DEST1 = "Stomp1";
    protected final static String DEST2 = "Stomp2";
    protected final static int NUM_MSGS = 1;
    protected static Logger LOG = LoggerFactory.getLogger(AMQTemplateTest.class);
    private BrokerService broker = null;


    /**
     * @throws java.lang.Exception
     */
    @Before
    public void setUp() throws Exception {
        LOG.info("Starting up");

        // start up broker instance
        try {
            Thread.currentThread().setContextClassLoader(AMQTemplateTest.class.getClassLoader());
            ClassPathXmlApplicationContext context = new ClassPathXmlApplicationContext(CONFIG_FILE);

            //resolve broker
            broker = (BrokerService) context.getBean("broker");

            //start the broker  
            if (!broker.isStarted()) {
                LOG.info("Broker broker not yet started. Kicking it off now.");
                broker.start();
            }
            else {
                LOG.info("Broker broker already started. Not kicking it off a second time.");
            }
        }
        catch(Exception e) {
            LOG.error("Error: " + e.getMessage());
            throw e;
        }
    }


    /**
     * @throws java.lang.Exception
     */
    @After
    public void tearDown() throws Exception {
        LOG.info("Shutting down");
        if (broker != null && broker.isStarted()) {
            LOG.info("Broker still running, stopping it now.");
            broker.stop();
        }
        else {
            LOG.info("Broker not running, nothing to shutdown.");
        }
    }


    /**
     * Tests for bug 
     * First populates a queue DEST1 with 100K Stomp messages so that broker's
     * MemoryPercentUsage >= 70%
     * 
     * Then sends one single Stomp msg to queue DEST2.
     * Then browses messages on DEST1.
     * Then browses messages on DEST2. This fails, no messages are returned.
     * Finally consumes the one msg from DEST2. This also fails, no message 
     * gets consumed.
     * 
     * Could not reproduce without Stomp messages on at least one destination.
     * 
     * @throws Exception
     */
    @Test
    public void testMessageDispatch() throws Exception {
        LOG.info("Testing for mesage dispatch.");

        // sent 150 msgs with 100K each to queue Stomp1 until MemoryLimit is reached
        Thread producerThread = new Thread(new StompProducer(150, DEST1, false, 0, 0));
        producerThread.setName("Producer for " + DEST1);
        producerThread.start();
        producerThread.join(5*60*1000);

        // sent 1 msg to queue Stomp2
        Thread producerThread2 = new Thread(new StompProducer(1, DEST2, false, 0, 0));
        producerThread2.setName("Producer for " + DEST2);
        producerThread2.start();
        producerThread2.join(10*1000);

        // browse DEST1 queue, if I skip this browse step, the unit test succeeds
        String errorMsg = "no messages to browse on queue " + DEST1 +
                " despite QueueSize=" + getQueueSize(DEST1);
        Assert.assertTrue(errorMsg, destinationHasMessages(DEST1));

        // browse DEST2 queue, will fail
        errorMsg = "no messages to browse on queue " + DEST2 +
                " despite QueueSize=" + getQueueSize(DEST2);
        Assert.assertTrue(errorMsg, destinationHasMessages(DEST2));

        // consume msg from DEST2 queue, will fail
        ExecutorService consumerThread = Executors.newSingleThreadExecutor();
        Future<Boolean> result = consumerThread.submit(new Consumer(DEST2, false));
        Assert.assertTrue("No message received from queue " + DEST2, result.get(20, TimeUnit.SECONDS));
    }

    /**
     * Creates queue browser and checks if browser contains any messages.
     * @param dest - JMS queue name to check
     * @return true if there are messages to browse on destingation, false if not.
     */
    protected boolean destinationHasMessages(String dest) {
        LOG.info("Trying to browse destination " + dest);

        Connection connection = null;
        Session session = null;
        try {
            ActiveMQConnectionFactory amq = new ActiveMQConnectionFactory(BROKERURL);
            connection = amq.createConnection();

            connection.setExceptionListener(new javax.jms.ExceptionListener() {
                    public void onException(javax.jms.JMSException e) {
                        e.printStackTrace();
                    }
            });
            connection.start();

            // Create a Session
            session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);

            // Create the destination (Topic or Queue)
            Destination  destination = null;
            destination = session.createQueue(dest);

            QueueBrowser browser = session.createBrowser((Queue)destination);
            Enumeration messages = browser.getEnumeration();
            return messages.hasMoreElements();
        } catch (Exception ex) {
            LOG.error(ex.toString(), ex);
        } finally {
            try {
                if (session != null)
                    session.close();
                if (connection != null)
                    connection.close();
            } catch (JMSException jmsx) {
                LOG.error("Error closing JMS resources: " + jmsx);
            }
        }
        return false;
    }

    /**
     * Retrieves JMX QueueSize attribute for a given queue name by connecting
     * to broker via JMX.
     * @param dest - JMS queue name to check
     * @return - QueueSize number as reported by JMX
     * @throws Exception
     */
    protected long getQueueSize(String dest) throws Exception {
        try {
            JMXServiceURL url = new JMXServiceURL("service:jmx:rmi:///jndi/rmi://:1099/jmxrmi");
            JMXConnector jmxc = JMXConnectorFactory.connect(url, null);
            MBeanServerConnection mbsc = jmxc.getMBeanServerConnection();
            Object tmp = mbsc.getAttribute(new ObjectName("org.apache.activemq:type=Broker,brokerName=localhost,destinationType=Queue,destinationName=" + dest), "QueueSize");
            long queueSize = ((Long) tmp).longValue();
            return queueSize;
        } catch (Exception ex) {
            LOG.error(ex.getMessage(), ex);
            throw ex;
        }
    }

    /**
     * Retrieves the broker's TotalMessageCount JMX attribute.
     * Opens a JMX connection to the broker to get the data.
     * @return - broker's TotalMessageCount as reported by JMX.
     * @throws Exception
     */
    protected long getTotalMessageCount() throws Exception {
        try {
            JMXServiceURL url = new JMXServiceURL("service:jmx:rmi:///jndi/rmi://:1099/jmxrmi");
            JMXConnector jmxc = JMXConnectorFactory.connect(url, null);
            MBeanServerConnection mbsc = jmxc.getMBeanServerConnection();
            Object tmp = mbsc.getAttribute(new ObjectName("org.apache.activemq:type=Broker,brokerName=localhost"), "TotalMessageCount");
            long queueSize = ((Long) tmp).longValue();
            return queueSize;
        } catch (Exception ex) {
            LOG.error(ex.getMessage(), ex);
            throw ex;
        }
    }
}
